from ..schemas import QueryBase
from typing import Optional, Literal, cast
from qdrant_client.conversions.common_types import Filter

def cleanse_query_for_search(query: Optional[QueryBase|dict], offset_type: Literal["int", "str"] = "int") -> Filter | None:
    if query is None: return None
    elif isinstance(query, dict):
        query = QueryBase(**query)
    if offset_type == "int":
        assert query.offset is None or isinstance(query.offset, int), "[Db] Offset must be int in Qdrant searching"
    elif offset_type == "str":
        assert query.offset is None or isinstance(query.offset, str), "[Db] Offset must be id (str) in Qdrant querying"
    assert query.id is None, "[Db] QueryBase's 'id' key should not being use for similarity search in qdrant"
    assert query.ids is None, "[Db] QueryBase's 'ids' key should not being use for similarity search in qdrant"
    assert query.order_by is None, "[Db] QueryBase's 'order_by' key should not being use for similarity search in qdrant"
    assert query.filters is None or isinstance(query.filters, dict), "[Db] Query.filters only accept dict of qdrant query"

    
    base = query.filters if query.filters is not None else {}

    # forbidden id and ids in filters
    if "id" in base: raise AssertionError(
        f"[Db] You should not include 'id' in your QueryBase.filters when searching Qdrant")
    if "ids" in base: raise AssertionError(
        f"[Db] You should not include 'ids' in your QueryBase.filters when searching Qdrant")
    return cast(Filter, base)



from typing import Generator
from pymongo import MongoClient
from qdrant_client import QdrantClient
from redis import Redis
from sqlmodel import Session

from ..session import Db


class ApiDbDeps:

    def __init__(self, db: Db):
        self.__db = db

    def get_redis(self) -> Generator[Redis | None, None, None]:
        r = None
        try:
            r = self.__db.get_redis(force_new_instance=True)
            yield r
        finally:
            self.__db.try_close(r)

    def get_mongo(self) -> Generator[MongoClient | None, None, None]:
        m = None
        try:
            m = self.__db.get_mongo(new_instance=True)
            yield m
        finally:
            self.__db.try_close(m)

    def get_write_sql(self) -> Generator[Session | None, None, None]:
        s = None
        try:
            s = self.__db.get_sql_write_client(new_instance=True)
            yield s
        finally:
            self.__db.try_close(s)

    def get_qdrant(self) -> Generator[QdrantClient | None, None, None]:
        q = None
        try:
            q = self.__db.get_qdrant_client(new_instance=True)
            yield q
        finally:
            self.__db.try_close(q)



from typing import Any, Generic, List, Mapping, Optional, Sequence, Type, TypeVar

from deprecated import deprecated
from motor.motor_asyncio import AsyncIOMotorClient
from pydantic import BaseModel
from pymongo import MongoClient
from pymongo.results import (
    DeleteResult,
    InsertManyResult,
    InsertOneResult,
    UpdateResult,
)

from ..schemas import QueryBase
from .mongo_util import (
    MaybeObjectId,
    build_object_id,
    cleanse_query_for_first,
    cleanse_query_for_list,
    cleanse_query_order_by,
)

ModelType = TypeVar("ModelType", bound=BaseModel)


class MongoCRUDBase(Generic[ModelType]):
    def __init__(
        self,
        model: Type[ModelType],
        collection: str,
        database: str,
        *,
        searchable: Optional[List[str]] = None,
    ):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD), in Mongo
        """
        self.model = model
        self.collection = collection
        self.searchable = searchable
        self.database = database

    def first(
        self,
        m: MongoClient,
        *,
        query: Optional[QueryBase | dict] = None,
        project: Optional[dict] = None,
        **kwargs,
    ) -> Optional[ModelType]:
        doc = self.first_doc(m, query=query, project=project, **kwargs)
        return self.model(**doc) if doc is not None else None

    async def afirst(
        self,
        m: AsyncIOMotorClient,
        *,
        query: Optional[QueryBase | dict] = None,
        project: Optional[dict] = None,
        **kwargs,
    ) -> Optional[ModelType]:
        """Asynchronously fetch the first document and parse it into a Pydantic model."""
        doc = await self.afirst_doc(m, query=query, project=project, **kwargs)
        return self.model(**doc) if doc else None

    def first_doc(
        self,
        m: MongoClient,
        *,
        query: Optional[QueryBase | dict] = None,
        project: Optional[dict] = None,
        **kwargs,
    ) -> Optional[dict]:
        doc: Optional[dict] = m[self.database][self.collection].find_one(
            {**cleanse_query_for_first(query)},
            project,
            **kwargs,
            sort=cleanse_query_order_by(query),
        )
        return doc

    async def afirst_doc(
        self,
        m: AsyncIOMotorClient,
        *,
        query: Optional[QueryBase | dict] = None,
        project: Optional[dict] = None,
        **kwargs,
    ) -> Optional[dict]:
        """Asynchronously fetch the first document matching the query."""
        doc = await m[self.database][self.collection].find_one(
            {**cleanse_query_for_first(query)},
            project,
            **kwargs,
            sort=cleanse_query_order_by(query),
        )
        return doc

    def get_multi(
        self,
        m: MongoClient,
        *,
        query: Optional[QueryBase | dict] = None,
        project: Optional[dict] = None,
    ) -> List[ModelType]:
        # assert query.offset is not None and query.limit is not None, 'offset and limit is require for listing query'
        if isinstance(query, dict):
            query = QueryBase(**query)
        _sort = cleanse_query_order_by(query)
        cursor = m[self.database][self.collection].find(
            {**cleanse_query_for_list(query)}, project
        )
        if _sort is not None and len(_sort) > 0:
            cursor = cursor.sort(_sort)
        if query is not None and query.offset is not None:
            if not isinstance(query.offset, int):
                raise ValueError(
                    f"[Db] Expected query offset to be type int, but got {query.offset}"
                )
            cursor = cursor.skip(int(query.offset))
        if query is not None and query.limit is not None:
            assert isinstance(
                query.offset, int
            ), f"[Db] Expected query limit to be type int, but got {query.limit}"
            cursor = cursor.limit(query.limit)
        docs: List[dict] = list(cursor)
        return [self.model(**doc) for doc in docs]

    async def aget_multi(
        self,
        m: AsyncIOMotorClient,
        *,
        query: Optional[QueryBase | dict] = None,
        project: Optional[dict] = None,
    ) -> List[ModelType]:
        """Asynchronously fetch multiple documents and parse them into Pydantic models."""
        if isinstance(query, dict):
            query = QueryBase(**query)
        _sort = cleanse_query_order_by(query)
        cursor = m[self.database][self.collection].find(
            {**cleanse_query_for_list(query)}, project
        )
        if _sort:
            cursor = cursor.sort(_sort)
        if query and query.offset is not None:
            if not isinstance(query.offset, int):
                raise ValueError(
                    f"[Db] Expected query offset to be type int, but got {query.offset}"
                )
            cursor = cursor.skip(query.offset)
        if query and query.limit is not None:
            cursor = cursor.limit(query.limit)
        docs = [doc async for doc in cursor]
        return [self.model(**doc) for doc in docs]

    def create(self, m: MongoClient, *, obj_in: ModelType) -> InsertOneResult:
        return m[self.database][self.collection].insert_one(obj_in.model_dump())

    async def acreate(
        self, m: AsyncIOMotorClient, *, obj_in: ModelType
    ) -> InsertOneResult:
        """Asynchronously create a single document."""
        return await m[self.database][self.collection].insert_one(obj_in.model_dump())

    def create_multi(
        self, m: MongoClient, *, obj_in_list: List[ModelType]
    ) -> InsertManyResult:
        return m[self.database][self.collection].insert_many(
            [obj_in.model_dump() for obj_in in obj_in_list]
        )

    async def acreate_multi(
        self, m: AsyncIOMotorClient, *, obj_in_list: List[ModelType]
    ) -> InsertManyResult:
        """Asynchronously create multiple documents."""
        return await m[self.database][self.collection].insert_many(
            [obj.model_dump() for obj in obj_in_list]
        )

    def update(
        self,
        m: MongoClient,
        *,
        _id: MaybeObjectId,
        obj_in: dict,
    ) -> UpdateResult:
        return m[self.database][self.collection].update_one(
            {"_id": build_object_id(_id)}, {"$set": obj_in}
        )

    async def aupdate(
        self, m: AsyncIOMotorClient, *, _id: MaybeObjectId, obj_in: dict
    ) -> UpdateResult:
        """Asynchronously update a document by ID."""
        return await m[self.database][self.collection].update_one(
            {"_id": build_object_id(_id)}, {"$set": obj_in}
        )

    @deprecated(
        reason="remove in next version, this method does nothing but setting a optional value in the mongo object dict, 'update' as True, which is inappropriate."
    )
    def hide(
        self,
        m: MongoClient,
        *,
        _id: MaybeObjectId,
    ) -> UpdateResult:
        return self.update(m, _id=_id, obj_in={"update": True})

    def delete(
        self,
        m: MongoClient,
        *,
        _id: MaybeObjectId,
    ) -> DeleteResult:
        return m[self.database][self.collection].delete_one(
            {"_id": build_object_id(_id)}
        )

    async def adelete(
        self, m: AsyncIOMotorClient, *, _id: MaybeObjectId
    ) -> DeleteResult:
        """Asynchronously delete a document by ID."""
        return await m[self.database][self.collection].delete_one(
            {"_id": build_object_id(_id)}
        )

    def delete_multi(
        self, m: MongoClient, *, query: Optional[QueryBase | dict]
    ) -> DeleteResult:
        if query is None:
            query_dict = {}
        else:
            if isinstance(query, dict):
                query = QueryBase(**query)
            query_dict: dict = query.model_dump(
                exclude_none=True, exclude_unset=True, exclude_defaults=True
            )
        if not ("filters" in query_dict and query_dict["filters"] is not None):
            raise ValueError(
                "[Db] You must set, and only set query.filters to delete_multi in mongodb"
            )
        return m[self.database][self.collection].delete_many(query_dict["filters"])

    async def adelete_multi(
        self, m: AsyncIOMotorClient, *, query: Optional[QueryBase | dict]
    ) -> DeleteResult:
        """Asynchronously delete multiple documents based on a query."""
        if query is None:
            query_dict = {}
        else:
            if isinstance(query, dict):
                query = QueryBase(**query)
            query_dict = query.model_dump(
                exclude_none=True, exclude_unset=True, exclude_defaults=True
            )
        if not ("filters" in query_dict and query_dict["filters"] is not None):
            raise ValueError(
                "[Db] You must set, and only set query.filters to adelete_multi in mongodb"
            )
        return await m[self.database][self.collection].delete_many(
            query_dict["filters"]
        )

    def first_by_id(
        self, m: MongoClient, *, _id: MaybeObjectId, **kwargs
    ) -> Optional[ModelType]:
        return self.first(
            m, query=QueryBase(id=_id if isinstance(_id, str) else str(_id)), **kwargs
        )  # convert type for inserted_id

    async def afirst_by_id(
        self, m: AsyncIOMotorClient, *, _id: MaybeObjectId, **kwargs
    ) -> Optional[ModelType]:
        """Asynchronously fetch the first document by ID and parse it into a Pydantic model."""
        return await self.afirst(
            m, query=QueryBase(id=_id if isinstance(_id, str) else str(_id)), **kwargs
        )

    def first_doc_by_id(
        self, m: MongoClient, *, _id: MaybeObjectId, **kwargs
    ) -> Optional[dict]:
        return self.first_doc(
            m, query=QueryBase(id=_id if isinstance(_id, str) else str(_id)), **kwargs
        )  # convert type for inserted_id

    async def afirst_doc_by_id(
        self, m: AsyncIOMotorClient, *, _id: MaybeObjectId, **kwargs
    ) -> Optional[dict]:
        """Asynchronously fetch the first document by ID."""
        return await self.afirst_doc(
            m, query=QueryBase(id=_id if isinstance(_id, str) else str(_id)), **kwargs
        )

    def aggregate(
        self,
        m: MongoClient,
        *,
        pipelines: Sequence[Mapping[str, Any]],
        model: Type[BaseModel],
        let: Optional[Mapping[str, Any]] = None,
    ) -> List[BaseModel]:
        docs = m[self.database][self.collection].aggregate(pipelines, let=let)
        return [model(**d) for d in docs]

    async def aaggregate(
        self,
        m: AsyncIOMotorClient,
        *,
        pipelines: Sequence[Mapping[str, Any]],
        model: Type[BaseModel],
        let: Optional[Mapping[str, Any]] = None,
    ) -> List[BaseModel]:
        """Asynchronously perform an aggregation query and parse results into Pydantic models."""
        cursor = m[self.database][self.collection].aggregate(pipelines, let=let)
        results = []
        async for doc in cursor:
            results.append(model(**doc))
        return results



from typing import Any, Generic, List, Optional, Type, TypeVar, Union, cast
from pydantic import BaseModel
from ..schemas import QueryBase
from .sql_util import parse_query_for_first, parse_query_for_list
from sqlmodel import Session, select
from sqlalchemy import func

ModelType = TypeVar("ModelType", bound=BaseModel)


class SqlCRUDBase(Generic[ModelType]):

    def __init__(self, model: Type[ModelType], table_name: str):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD) in SQL

        **Parameters**
        * `model`: A SQLAlchemy model class
        """
        self.model = model
        self.model.__table__.name = table_name # type: ignore
        self.model.__tablename__ = table_name # type: ignore
        self.table_name = table_name

    def first(self, s: Session, query: Optional[QueryBase] = None) -> Optional[ModelType]:
        parsed = parse_query_for_first(self.model, query=query)
        if query is not None:
            sel = select(self.model).order_by(query.order_by)
        else:
            sel = select(self.model)
        statement = sel.limit(1)  # only the first one
        if parsed.where is not None: statement = statement.where(parsed.where)
        results = s.exec(statement)
        return results.one_or_none()

    def get_multi(
            self, s: Session, *, query: QueryBase | dict | None
    ) -> List[ModelType]:
        if isinstance(query, dict):
            query = QueryBase(**query)
        # assert query.offset is not None and query.limit is not None, '[Db] offset and limit is require for listing query'
        parsed = parse_query_for_list(self.model, query=query)
        statement = select(self.model)
        if query is not None and query.order_by is not None:
            statement = statement.order_by(query.order_by)
        if query is not None and query.limit is not None:
            statement = statement.limit(query.limit)
        if query is not None and query.offset is not None:
            assert isinstance(query.offset, int), f"[Db] Expected int or None type offset in sql query, but got {query.offset}"
            statement = statement.offset(query.offset)
        if parsed.where is not None: statement = statement.where(parsed.where)
        results = s.exec(statement)
        return list(results.all())

    def count(self, s: Session, query: Optional[QueryBase] = None) -> int:
        parsed = parse_query_for_list(self.model, query=query)
        statement = select(func.count(self.model.id))  # type: ignore # only the first one
        if parsed.where is not None: statement = statement.where(parsed.where)
        count = s.exec(statement).one()
        return count

    def create(self, s: Session, *, obj_in: ModelType):
        obj_in = self.pre_commit_create(s, obj_in=obj_in)
        s.commit()
        s.refresh(obj_in)
        return obj_in

    def pre_commit_create(self, s: Session, *, obj_in: ModelType):
        s.add(obj_in)
        return obj_in

    def create_multi(self, s: Session, *, obj_in_list: List[ModelType]):
        if len(obj_in_list) == 0: return []
        for obj_in in obj_in_list:
            s.add(obj_in)
        s.commit()
        for obj_in in obj_in_list:
            s.refresh(obj_in)
        return obj_in_list

    def update(
            self, s: Session, *, _id: str, obj_in: dict
    ) -> ModelType:
        obj = self.pre_commit_update(s, _id=_id, obj_in=obj_in)
        s.commit()
        s.refresh(obj)
        return obj

    def pre_commit_update(
            self, s: Session, *, _id: str, obj_in: dict,
    ) -> ModelType:
        if not isinstance(obj_in, dict):
            # try to convert to dict from pydantic object
            assert hasattr(obj_in, "model_dump") and callable(
                obj_in.model_dump), f"[Db] Expected dict or pydantic object, but got type: {type(obj_in)}"
            obj_in = obj_in.model_dump(exclude={"id", "created"})

        obj = self.first_by_id(s, _id=_id)
        if obj is None:
            raise ValueError("Failed to update due to object not found")
        for k, v in obj_in.items():
            # print("k, v ", k, v)
            # eval(f"obj.{k}=v")
            if hasattr(obj, k):
                setattr(obj, k, v)
            # setattr(obj, k, v)
        s.add(obj)
        return obj

    def delete(
            self, s: Session, *, _id: str,
    ):
        obj = self.first_by_id(s, _id=_id)
        if obj is None:
            return None
        s.delete(obj)
        s.commit()

    def delete_multi(self, s: Session, *, obj_in_list: List[Union[str, ModelType]]):
        if len(obj_in_list) == 0: return
        ids = [item for item in obj_in_list if isinstance(item, str)]
        objects = [item for item in obj_in_list if not isinstance(item, str)]

        if len(objects) > 0:
            for obj in objects:
                s.delete(obj)
            s.commit()
        for _id in ids:
            self.delete(s, _id=_id)

    def first_by_id(self, s: Session, *, _id: str) -> Optional[ModelType]:
        return self.first(s, query=QueryBase(id=_id))
